from __future__ import annotations
import builtins as __builtins__
import cupy as cp
import hashlib as hashlib
import io as io
import numpy as np
import os as os
import typing
__all__ = ['cp', 'hashlib', 'io', 'load_emap', 'np', 'os', 'save', 'script_dir']
def load_emap(emap_path: str | None = None) -> cp.ndarray:
    ...
def save(array: np.ndarray, dest_path: str | None = None):
    ...
__test__: dict = {}
script_dir: str = 'Z:\\Repositories_mia\\visagene\\src\\visagene'
